import asyncio
from types import TracebackType
from typing import Optional, Union, Type, List

from .aio import simple_cache_client as aio

from ._async_utils import wait_for_coroutine
from .cache_operation_types import (
    CacheGetResponse,
    CacheSetResponse,
    CreateCacheResponse,
    DeleteCacheResponse,
    ListCachesResponse,
    CacheMultiGetResponse,
    CacheMultiGetOperation,
    CacheMultiSetOperation,
    CacheMultiSetResponse,
    CacheMultiSetFailureResponse,
    CacheMultiGetFailureResponse,
)

from ._utilities._data_validation import _validate_request_timeout


class SimpleCacheClient:
    def __init__(
        self,
        auth_token: str,
        default_ttl_seconds: int,
        data_client_operation_timeout_ms: Optional[int],
    ):
        try:
            # If the synchronous client is used inside an async application,
            # use the event loop it's running within.
            loop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
        except RuntimeError:
            # Currently, we rely on asyncio's module-wide event loop due to the
            # way the grpc stubs we've got are hiding the _loop parameter.
            # If a separate loop is required, e.g., so you can run Simple Cache
            # on a background thread, you'll want to open an issue.
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        self._loop = loop

        self._momento_async_client = aio.SimpleCacheClient(
            auth_token=auth_token,
            default_ttl_seconds=default_ttl_seconds,
            data_client_operation_timeout_ms=data_client_operation_timeout_ms,
        )

    def __enter__(self) -> "SimpleCacheClient":
        wait_for_coroutine(self._loop, self._momento_async_client.__aenter__())
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        wait_for_coroutine(
            self._loop,
            self._momento_async_client.__aexit__(exc_type, exc_value, traceback),
        )

    def create_cache(self, cache_name: str) -> CreateCacheResponse:
        """Creates a new cache in your Momento account.

        Args:
            cache_name: String used to create cache.

        Returns:
            CreateCacheResponse

        Raises:
            InvalidArgumentError: If provided cache_name is None.
            BadRequestError: If the cache name provided doesn't follow the naming conventions
            AlreadyExistsError: If cache with the given name already exists.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            ClientSdkError: For any SDK checks that fail.
        """
        coroutine = self._momento_async_client.create_cache(cache_name)
        return wait_for_coroutine(self._loop, coroutine)

    def delete_cache(self, cache_name: str) -> DeleteCacheResponse:
        """Deletes a cache and all the items within it.

        Args:
            cache_name: String cache name to delete.

        Returns:
            DeleteCacheResponse

        Raises:
            InvalidArgumentError: If provided cache_name is None.
            BadRequestError: If the cache name provided doesn't follow the naming conventions
            NotFoundError: If an attempt is made to delete a MomentoCache that doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            ClientSdkError: For any SDK checks that fail.
        """
        coroutine = self._momento_async_client.delete_cache(cache_name)
        return wait_for_coroutine(self._loop, coroutine)

    def list_caches(self, next_token: Optional[str] = None) -> ListCachesResponse:
        """Lists all caches.

        Args:
            next_token: Token to continue paginating through the list. It's used to handle large paginated lists.

        Returns:
            ListCachesResponse

        Raises:
            AuthenticationError: If the provided Momento Auth Token is invalid.
        """
        coroutine = self._momento_async_client.list_caches(next_token)
        return wait_for_coroutine(self._loop, coroutine)

    def set(
        self,
        cache_name: str,
        key: str,
        value: Union[str, bytes],
        ttl_seconds: Optional[int] = None,
    ) -> CacheSetResponse:
        """Stores an item in cache

        Args:
            cache_name: Name of the cache to store the item in.
            key (string or bytes): The key to be used to store item.
            value (string or bytes): The value to be stored.
            ttl_seconds (Optional): Time to live in cache in seconds. If not provided, then default TTL for the cache
                client instance is used.

        Returns:
            CacheSetResponse

        Raises:
            InvalidArgumentError: If validation fails for the provided method arguments.
            BadRequestError: If the provided inputs are rejected by server because they are invalid
            NotFoundError: If the cache with the given name doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            InternalServerError: If server encountered an unknown error while trying to store the item.
        """
        coroutine = self._momento_async_client.set(cache_name, key, value, ttl_seconds)
        return wait_for_coroutine(self._loop, coroutine)

    def multi_set(
        self,
        cache_name: str,
        ops: Union[List[CacheMultiSetOperation], List[CacheMultiSetFailureResponse]],
    ) -> CacheMultiSetResponse:
        """Executes a list of passed Set operations in parallel.

        Args:
            cache_name: Name of the cache to store the item in.
            ops: (List[CacheMultiSetOperation]): List of set operations to execute.

        Returns:
            CacheMultiGetResponse

        Raises:
            InvalidArgumentError: If validation fails for the provided method arguments.
            BadRequestError: If the provided inputs are rejected by server because they are invalid
            NotFoundError: If the cache with the given name doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            InternalServerError: If server encountered an unknown error while trying to retrieve the item.
        """
        coroutine = self._momento_async_client.multi_set(cache_name, ops)
        return wait_for_coroutine(self._loop, coroutine)

    def get(self, cache_name: str, key: str) -> CacheGetResponse:
        """Retrieve an item from the cache

        Args:
            cache_name: Name of the cache to get the item from
            key (string or bytes): The key to be used to retrieve the item.

        Returns:
            CacheGetResponse

        Raises:
            InvalidArgumentError: If validation fails for the provided method arguments.
            BadRequestError: If the provided inputs are rejected by server because they are invalid
            NotFoundError: If the cache with the given name doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            InternalServerError: If server encountered an unknown error while trying to retrieve the item.
        """
        coroutine = self._momento_async_client.get(cache_name, key)
        return wait_for_coroutine(self._loop, coroutine)

    def multi_get(
        self,
        cache_name: str,
        ops: Union[List[CacheMultiGetOperation], List[CacheMultiGetFailureResponse]],
    ) -> CacheMultiGetResponse:
        """Executes a list of passed Get operations in parallel.

        Args:
            cache_name: Name of the cache to get the item from.
            ops: (Union[List[CacheMultiGetOperation], List[CacheMultiGetFailureResponse]]): List of get operations to
                execute.

        Returns:
            CacheMultiGetResponse

        Raises:
            InvalidArgumentError: If validation fails for the provided method arguments.
            BadRequestError: If the provided inputs are rejected by server because they are invalid
            NotFoundError: If the cache with the given name doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            InternalServerError: If server encountered an unknown error while trying to retrieve the item.
        """
        coroutine = self._momento_async_client.multi_get(cache_name, ops)
        return wait_for_coroutine(self._loop, coroutine)


def init(
    auth_token: str,
    item_default_ttl_seconds: int,
    request_timeout_ms: Optional[int] = None,
) -> SimpleCacheClient:
    """Creates a SimpleCacheClient

    Args:
        auth_token: Momento Token to authenticate the requests with Simple Cache Service
        item_default_ttl_seconds: A default Time To Live in seconds for cache objects created by this client. It is
            possible to override this setting when calling the set method.
        request_timeout_ms: An optional timeout in milliseconds to allow for Get and Set operations to complete.
            Defaults to 5 seconds. The request will be terminated if it takes longer than this value and will result
            in TimeoutError.
    Returns:
        SimpleCacheClient
    Raises:
        IllegalArgumentError: If method arguments fail validations
    """
    _validate_request_timeout(request_timeout_ms)
    return SimpleCacheClient(auth_token, item_default_ttl_seconds, request_timeout_ms)



from typing import Optional, Union, List

from grpc.aio import Metadata

from .. import errors
from ..cache_operation_types import (
    CacheMultiSetOperation,
    CacheMultiGetOperation,
    CacheMultiSetFailureResponse,
    CacheMultiGetFailureResponse,
)


def _make_metadata(cache_name: str) -> Metadata:
    return Metadata(("cache", cache_name))  # type: ignore[misc]


def _validate_multi_op_list(
    values: Union[
        List[CacheMultiSetOperation],
        List[CacheMultiGetOperation],
        List[CacheMultiSetFailureResponse],
        List[CacheMultiGetFailureResponse],
    ]
) -> None:
    if len(values) < 1:
        raise errors.InvalidArgumentError("empty op list passed to multi operation")


def _validate_cache_name(cache_name: str) -> None:
    if cache_name is None or not isinstance(cache_name, str):
        raise errors.InvalidArgumentError("Cache name must be a non-empty string")


def _as_bytes(data: Union[str, bytes], error_message: str) -> bytes:
    if isinstance(data, str):
        return data.encode("utf-8")
    if isinstance(data, bytes):
        return data
    raise errors.InvalidArgumentError(error_message + str(type(data)))


def _validate_ttl(ttl_seconds: int) -> None:
    if not isinstance(ttl_seconds, int) or ttl_seconds < 0:
        raise errors.InvalidArgumentError("TTL Seconds must be a non-negative integer")


def _validate_request_timeout(request_timeout_ms: Optional[int]) -> None:
    if request_timeout_ms is None:
        return
    if not isinstance(request_timeout_ms, int) or request_timeout_ms <= 0:
        raise errors.InvalidArgumentError("Request timeout must be greater than zero.")



import asyncio
import functools
from typing import Union, Optional, List, Any
from momento_wire_types.cacheclient_pb2 import _GetRequest
from momento_wire_types.cacheclient_pb2 import _SetRequest

from .. import cache_operation_types as cache_sdk_ops
from .. import _cache_service_errors_converter
from .. import _momento_logger
from . import _scs_grpc_manager
from .._utilities._data_validation import (
    _as_bytes,
    _validate_ttl,
    _make_metadata,
    _validate_cache_name,
    _validate_multi_op_list,
)

_DEFAULT_DEADLINE_SECONDS = 5.0  # 5 seconds


class _ScsDataClient:
    """Internal"""

    def __init__(
        self,
        auth_token: str,
        endpoint: str,
        default_ttl_seconds: int,
        operation_timeout_ms: Optional[int],
    ):
        self._default_deadline_seconds = (
            _DEFAULT_DEADLINE_SECONDS
            if not operation_timeout_ms
            else operation_timeout_ms / 1000.0
        )
        self._grpc_manager = _scs_grpc_manager._DataGrpcManager(auth_token, endpoint)
        _validate_ttl(default_ttl_seconds)
        self._default_ttlSeconds = default_ttl_seconds

    async def set(
        self,
        cache_name: str,
        key: Union[str, bytes],
        value: Union[str, bytes],
        ttl_seconds: Optional[int],
    ) -> cache_sdk_ops.CacheSetResponse:
        _validate_cache_name(cache_name)
        try:
            _momento_logger.debug(f"Issuing a set request with key {str(key)}")
            item_ttl_seconds = (
                self._default_ttlSeconds if ttl_seconds is None else ttl_seconds
            )
            _validate_ttl(item_ttl_seconds)
            set_request = _SetRequest()
            set_request.cache_key = _as_bytes(key, "Unsupported type for key: ")
            set_request.cache_body = _as_bytes(value, "Unsupported type for value: ")
            set_request.ttl_milliseconds = item_ttl_seconds * 1000
            response = await self._grpc_manager.async_stub().Set(
                set_request,
                metadata=_make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            _momento_logger.debug(f"Set succeeded for key: {str(key)}")
            return cache_sdk_ops.CacheSetResponse(
                response, set_request.cache_key, set_request.cache_body
            )
        except Exception as e:
            _momento_logger.debug(f"Set failed for {str(key)} with response: {e}")
            raise _cache_service_errors_converter.convert(e)

    async def multi_set(
        self,
        cache_name: str,
        set_operations: Union[
            List[cache_sdk_ops.CacheMultiSetOperation],
            List[cache_sdk_ops.CacheMultiSetFailureResponse],
        ],
    ) -> cache_sdk_ops.CacheMultiSetResponse:

        _validate_multi_op_list(set_operations)
        _validate_cache_name(cache_name)

        # Will hold which tasks have succeeded or failed to return to caller
        failed_ops: List[cache_sdk_ops.CacheMultiSetFailureResponse] = []
        successful_ops: List[cache_sdk_ops.CacheSetResponse] = []

        try:
            requests: List[_SetRequest] = list()
            request_promises = list()
            for op in set_operations:
                item_ttl_seconds = (
                    self._default_ttlSeconds
                    if op.ttl_seconds is None
                    else op.ttl_seconds
                )
                _validate_ttl(item_ttl_seconds)
                set_request = _SetRequest()
                set_request.cache_key = _as_bytes(op.key, "Unsupported type for key: ")
                set_request.cache_body = _as_bytes(
                    op.value, "Unsupported type for value: "
                )
                set_request.ttl_milliseconds = item_ttl_seconds * 1000

                requests.append(set_request)
                request_promises.append(
                    self._grpc_manager.async_stub().Set(
                        set_request,
                        metadata=_make_metadata(cache_name),
                        timeout=self._default_deadline_seconds,
                    )
                )

            results = await asyncio.gather(
                *request_promises,
                return_exceptions=True,
                # When set to True exceptions are treated the same as successful results, and aggregated in the
                # result list. We want to try and make sure all requests have chance to finish.
            )

            for request, result in zip(requests, results):
                if isinstance(result, Exception):
                    _momento_logger.debug(
                        f"multi-set sub command failed with "
                        f"error: {result} "
                        f"key: {str(request.cache_key)}"
                    )
                    failed_ops.append(
                        cache_sdk_ops.CacheMultiSetFailureResponse(
                            key=request.cache_key,
                            value=request.cache_body,
                            ttl_seconds=int(request.ttl_milliseconds / 1000),
                            failure=_cache_service_errors_converter.convert(result),
                        )
                    )
                else:
                    successful_ops.append(
                        cache_sdk_ops.CacheSetResponse(
                            result, request.cache_key, request.cache_body
                        )
                    )
            _momento_logger.debug(
                f"multi-set succeeded "
                f"successful_op_count={len(successful_ops)} "
                f"failed_op_count={len(failed_ops)}"
            )
            return cache_sdk_ops.CacheMultiSetResponse(
                successful_responses=successful_ops,
                failed_responses=failed_ops,
            )
        except Exception as e:
            _momento_logger.debug(f"multi-set failed with error: {e}")
            # re-raise any error caught here is fatal error with overall handling of request objects
            raise _cache_service_errors_converter.convert(e)

    async def get(
        self, cache_name: str, key: Union[str, bytes]
    ) -> cache_sdk_ops.CacheGetResponse:

        _validate_cache_name(cache_name)
        try:
            _momento_logger.debug(f"Issuing a get request with key {str(key)}")
            get_request = _GetRequest()
            get_request.cache_key = _as_bytes(key, "Unsupported type for key: ")
            response = await self._grpc_manager.async_stub().Get(
                get_request,
                metadata=_make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            _momento_logger.debug(f"Received a get response for {str(key)}")
            return cache_sdk_ops.CacheGetResponse(response)
        except Exception as e:
            _momento_logger.debug(f"Get failed for {str(key)} with response: {e}")
            raise _cache_service_errors_converter.convert(e)

    async def multi_get(
        self,
        cache_name: str,
        get_operations: Union[
            List[cache_sdk_ops.CacheMultiGetOperation],
            List[cache_sdk_ops.CacheMultiGetFailureResponse],
        ],
    ) -> cache_sdk_ops.CacheMultiGetResponse:
        _validate_multi_op_list(get_operations)
        _validate_cache_name(cache_name)
        try:
            request_promises = list()
            requests: List[_GetRequest] = list()
            for op in get_operations:
                get_request = _GetRequest()
                get_request.cache_key = _as_bytes(op.key, "Unsupported type for key: ")
                requests.append(get_request)
                request_promises.append(
                    self._grpc_manager.async_stub().Get(
                        get_request,
                        metadata=_make_metadata(cache_name),
                        timeout=self._default_deadline_seconds,
                    )
                )

            success_responses: List[cache_sdk_ops.CacheGetResponse] = []
            failed_responses: List[cache_sdk_ops.CacheMultiGetFailureResponse] = []
            results = await asyncio.gather(
                *request_promises,
                return_exceptions=True,
                # When set to True exceptions are treated the same as successful results, and aggregated in the
                # result list. We want to try and make sure all requests have chance to finish.
            )
            for request, result in zip(requests, results):
                if isinstance(result, Exception):
                    failed_responses.append(
                        cache_sdk_ops.CacheMultiGetFailureResponse(
                            key=request.cache_key,
                            failure=_cache_service_errors_converter.convert(result),
                        )
                    )
                else:
                    success_responses.append(cache_sdk_ops.CacheGetResponse(result))

            _momento_logger.debug(
                f"multi_get succeeded "
                f"success_count={len(success_responses)} "
                f"failure_count={len(failed_responses)} "
            )
            return cache_sdk_ops.CacheMultiGetResponse(
                successful_responses=success_responses,
                failed_responses=failed_responses,
            )
        except Exception as e:
            _momento_logger.debug(f"multi_get failed with response: {e}")
            raise _cache_service_errors_converter.convert(e)

    async def close(self) -> None:
        await self._grpc_manager.close()



from typing import Callable, Union, List

import grpc
from grpc.aio import Metadata


class Header:
    once_only_headers = ["agent"]

    def __init__(self, name: str, value: str):
        self.name = name
        self.value = value


class AddHeaderClientInterceptor(grpc.aio.UnaryUnaryClientInterceptor):
    are_only_once_headers_sent = False

    def __init__(self, headers: List[Header]):
        self._headers_to_add_once = list(
            filter(lambda header: header.name in header.once_only_headers, headers)
        )
        self.headers_to_add_every_time = list(
            filter(lambda header: header.name not in header.once_only_headers, headers)
        )

    async def intercept_unary_unary(
        self,
        continuation: Callable[
            [grpc.aio._interceptor.ClientCallDetails, grpc.aio._typing.RequestType],
            grpc.aio._call.UnaryUnaryCall,
        ],
        client_call_details: grpc.aio._interceptor.ClientCallDetails,
        request: grpc.aio._typing.RequestType,
    ) -> Union[grpc.aio._call.UnaryUnaryCall, grpc.aio._typing.ResponseType]:
        if client_call_details.metadata is None:
            client_call_details.metadata = Metadata()
        for header in self.headers_to_add_every_time:
            client_call_details.metadata.add(header.name, header.value)
        if AddHeaderClientInterceptor.are_only_once_headers_sent == False:
            for header in self._headers_to_add_once:
                client_call_details.metadata.add(header.name, header.value)
            AddHeaderClientInterceptor.are_only_once_headers_sent = True

        return await continuation(client_call_details, request)
